package com.dsk.flink.dsc.common.function;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.map.MapUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.dsk.flink.dsc.utils.EnvProperties;
import org.apache.flink.api.java.tuple.Tuple3;
import org.apache.flink.api.java.tuple.Tuple6;
import org.apache.flink.streaming.api.functions.ProcessFunction;
import org.apache.flink.util.Collector;
import org.apache.flink.util.OutputTag;

import java.util.*;
/**
 * 重构代码
 * @author lww
 * @date 2025-01-14
 */
public class MysqlDataTransferFunction extends ProcessFunction<JSONObject, Tuple3<String,String,Long>> {

    private static final Map<String,Integer> STR_SQL_TYPE;
    private final EnvProperties dbInfoMap;
    private final OutputTag<Tuple6<String,String,String,String,String,Long>> logSlideTag;

    static {
        STR_SQL_TYPE = MapUtil.newHashMap();
        STR_SQL_TYPE.put("VARCHAR",1);
        STR_SQL_TYPE.put("CHAR",1);
        STR_SQL_TYPE.put("TINYBLOB",1);
        STR_SQL_TYPE.put("BLOB",1);
        STR_SQL_TYPE.put("MEDIUMBLOB",1);
        STR_SQL_TYPE.put("LONGBLOB",1);
        STR_SQL_TYPE.put("TINYTEXT",1);
        STR_SQL_TYPE.put("TEXT",1);
        STR_SQL_TYPE.put("MEDIUMTEXT",1);
        STR_SQL_TYPE.put("LONGTEXT",1);
        STR_SQL_TYPE.put("TIME",1);
        STR_SQL_TYPE.put("TIMESTAMP",1);
        STR_SQL_TYPE.put("JSON",1);
    }

    public MysqlDataTransferFunction(EnvProperties envProps, OutputTag<Tuple6<String,String,String,String,String,Long>> logSlideTag) {
        this.dbInfoMap = envProps;
        this.logSlideTag = logSlideTag;
    }

    @Override
    public void processElement(JSONObject value, Context ctx, Collector<Tuple3<String, String, Long>> out) throws Exception {
        //返回数据集合
        String type = value.getString("type");
        JSONObject mysqlType = value.getJSONObject("mysqlType");
        String table = value.getString("table");
        JSONArray pkNames = value.getJSONArray("pkNames");
        Set<String> pkNameSet = new HashSet<>();
        long ts = value.getLong("ts") == null ? System.currentTimeMillis() : value.getLong("ts");
        if(CollUtil.isNotEmpty(pkNames)){
            pkNames.forEach(name -> pkNameSet.add(String.valueOf(name)));
        }
        String excueteSql;
        JSONObject dataObj = value.getJSONArray("data").getJSONObject(0);

        Boolean logicalDelete = MapUtil.getBool(dbInfoMap, "logical_delete", false);
        if(logicalDelete){
            mysqlType.put("is_del", "int");
            dataObj.put("is_del", "DELETE".equals(type) ? 1 : 0);
        }
        StringBuilder groupKeyBuilder = new StringBuilder(table);
        StringBuilder pkColumns = new StringBuilder();
        StringBuilder pkColumnVals = new StringBuilder();
        for (String pk : pkNameSet) {
            String pkValue = getValueString(dataObj, pk, mysqlType.getString(pk));
            groupKeyBuilder.append("-").append(pkValue);
            pkColumns.append(pk).append(",");
            pkColumnVals.append(pkValue).append("-");
        }
        String groupKey = groupKeyBuilder.toString();

        //添加分表参数
        String shardingRule = dataObj.getString("shardingRule");
        if (StrUtil.isNotBlank(shardingRule)) {
            Map<String,Object> map = JSON.parseObject(shardingRule, Map.class);
            String strategy = MapUtil.getStr(map, "strategy");
            Map<String,Object> strategyMap = JSON.parseObject(strategy, Map.class);
            int i = MapUtil.getInt(strategyMap, "sharding-count");
            if (i > 1){
                String str = MapUtil.getStr(strategyMap, "sharding-column");
                Integer val = dataObj.getInteger(str);
                val = val == null ? 1 : val;
                table = table.concat("_").concat(String.valueOf(val % i));
            }
        }
        if("INSERT".equals(type) || "UPDATE".equals(type)){
            excueteSql = tranferInsertSql(table,dataObj,mysqlType);
        } else {
            excueteSql = transferDeleteSql(table,dataObj,mysqlType,pkNameSet);
        }
        if (MapUtil.getBool(dbInfoMap, "log_enable", false)){
            ctx.output(logSlideTag,buildLogData(type, table, pkColumns, pkColumnVals, ts, value.toJSONString()));
        }
        out.collect(Tuple3.of(excueteSql,groupKey,ts));
    }

    private static Tuple6<String,String,String,String,String,Long> buildLogData(String type, String table, StringBuilder pkColumns, StringBuilder pkValues, long ts, String dataJsonStr) {
        if (pkColumns.length() > 0) {
            pkColumns.setLength(pkColumns.length()-1);
            pkValues.setLength(pkValues.length()-1);
        }
        return Tuple6.of(table,type, pkColumns.toString(), pkValues.toString().replace("'",""),dataJsonStr.replace("\\","\\\\").replace("'", "\\'"),ts);
    }

    private static String tranferInsertSql(String table, JSONObject dataObj, JSONObject mysqlType) {
        Set<String> columnSet = mysqlType.keySet();
        StringBuilder sb = new StringBuilder("REPLACE INTO ").append(table).append(" (");
        List<String> valueList = new ArrayList<>();
        for (String s : columnSet) {
            sb.append("`").append(s).append("`,");
            valueList.add(getValueString(dataObj, s, mysqlType.getString(s)));
        }
        sb.setLength(sb.length() - 1);
        sb.append(") values (");
        sb.append(String.join(",", valueList));
        sb.append(")");
        return sb.toString();
    }

    private static String transferDeleteSql(String table, JSONObject dataObj, JSONObject mysqlType, Set<String> pkNameSet) {
        StringBuilder whereClauseBuilder = new StringBuilder();
        for (String pk : pkNameSet) {
            if (whereClauseBuilder.length() > 0) {
                whereClauseBuilder.append(" and ");
            }
            whereClauseBuilder.append(pk).append(" = ").append(getValueString(dataObj, pk, mysqlType.getString(pk)));
        }
        return String.format("DELETE FROM %s WHERE %s",table, whereClauseBuilder);
    }

    private static String getValueString(JSONObject dataObj,String columnKey,String mysqlType){
        if(null == dataObj.get(columnKey)){
            return "null";
        }
        String upperCase = mysqlType.toUpperCase();
        //需要处理成字符串加引号的类型
        if(STR_SQL_TYPE.containsKey(upperCase)){
            return String.format("'%s'", dataObj.getString(columnKey).replace("\\","\\\\").replace("'", "\\'") );
        }

        //时间字段处理
        if("DATE".equals(upperCase) || "DATETIME".equals(upperCase)){
            String date = "DATETIME".equals(upperCase) ? DateUtil.format(dataObj.getDate(columnKey),"yyyy-MM-dd HH:mm:ss") : DateUtil.format(dataObj.getDate(columnKey),"yyyy-MM-dd");
            return String.format("\"%s\"",date);
        }

        return dataObj.getString(columnKey);
    }
}
