package com.dsk.flink.dsc.common.sink;

import cn.hutool.core.lang.Snowflake;
import cn.hutool.core.util.IdUtil;
import cn.hutool.core.util.RandomUtil;
import com.alibaba.druid.pool.DruidDataSource;
import com.dsk.flink.dsc.common.dto.SqlErrorLog;
import com.dsk.flink.dsc.utils.EnvProperties;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.sink.RichSinkFunction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 * @author lww
 * @date 2025-01-14
 */
public class MysqlDataTransferSinkBatch extends RichSinkFunction<String> {

    static Logger logger = LoggerFactory.getLogger(MysqlDataTransferSink.class);
    EnvProperties envProps;
    private static transient ExecutorService executorService;
    private static transient DruidDataSource dataSource;
    private static final int BATCH_SIZE = 1000;
    private static final int FLUSH_INTERVAL = 500;
    private ArrayBlockingQueue<String> sqlBatch = new ArrayBlockingQueue<>(BATCH_SIZE * 2);
    private static transient ScheduledExecutorService scheduledExecutorService;
    private AtomicBoolean flushing = new AtomicBoolean(false);

    public MysqlDataTransferSinkBatch(EnvProperties envProps) {
        this.envProps = envProps;
    }

    @Override
    public void open(Configuration parameters) throws Exception {
        executorService = new ThreadPoolExecutor(10, 10, 20, TimeUnit.MINUTES, new ArrayBlockingQueue<>(1000));
        // 初始化获取配置
        String configTidbUrl = String.format(envProps.getDb_url(), envProps.getDb_host(), envProps.getDb_port(), envProps.getDb_database());
        dataSource = new DruidDataSource();
        dataSource.setDriverClassName("com.mysql.cj.jdbc.Driver");
        dataSource.setUsername(envProps.getDb_username());
        dataSource.setPassword(envProps.getDb_password());
        dataSource.setUrl(configTidbUrl);
        dataSource.setMaxActive(30);
        dataSource.setInitialSize(20);
        dataSource.setTestWhileIdle(true);
        dataSource.setMaxWait(20000);
        dataSource.setValidationQuery("select 1");
        scheduledExecutorService = new ScheduledThreadPoolExecutor(1);
        scheduledExecutorService.scheduleAtFixedRate(this::flush, FLUSH_INTERVAL, FLUSH_INTERVAL, TimeUnit.MILLISECONDS);
    }

    @Override
    public void close() throws Exception {
        executorService.shutdown();
        scheduledExecutorService.shutdown();
        dataSource.close();
    }

    @Override
    public void invoke(String value, Context context) throws Exception {
        sqlBatch.put(value);
        if (sqlBatch.size() >= BATCH_SIZE) {
            flush();
        }
    }

    private void flush() {
        if (flushing.compareAndSet(false, true)) {
            try {
                if (sqlBatch.isEmpty()) {
                    flushing.set(false);
                    return;
                }
                List<String> batch = new ArrayList<>(1334);
                sqlBatch.drainTo(batch, BATCH_SIZE);
                if (batch.isEmpty()) {
                    flushing.set(false);
                    return;
                }
                executorService.execute(() -> {
                    try (Connection connection = dataSource.getConnection()) {
                        connection.setAutoCommit(false);
                        try (Statement pt = connection.createStatement()) {
                            for (String sql : batch) {
                                pt.addBatch(sql);
                            }
                            pt.executeBatch();
                            connection.commit();
                        } catch (Exception e) {
                            try {
                                connection.rollback();
                            } catch (Exception ex) {
                                logger.error("事务回滚异常", ex);
                            }
                            logger.error("------错误时间:{}-----,sql:{}--------异常:{}", new Date(), String.join(";", batch), e);
                            SqlErrorLog errorLog = new SqlErrorLog(new Date(), String.join(";", batch), e.getMessage());
                            writeErrLogDb(errorLog);
                        }
                    } catch (Exception e) {
                        logger.error("获取数据库连接异常", e);
                    } finally {
                        flushing.set(false);
                    }
                });
            } catch (Exception e) {
                flushing.set(false);
                logger.error("Flush 操作异常", e);
            }
        }
    }


    private void writeErrLogDb(SqlErrorLog errorLog) {
        Snowflake snowflake = IdUtil.getSnowflake(RandomUtil.randomInt(31), RandomUtil.randomInt(31));
        String sql = "insert dsc_err_log (id,error_time, error_sql, error_msg) values (?,?,?,?)";
        try (Connection conn = dataSource.getConnection();
             PreparedStatement pt = conn.prepareStatement(sql)) {
            pt.setLong(1, snowflake.nextId());
            pt.setObject(2, errorLog.getErrorTime());
            pt.setString(3, errorLog.getSql());
            pt.setString(4, errorLog.getError());
            pt.execute();
        } catch (Exception e) {
            logger.error("错误日志保存异常", e);
        }
    }

}